@@ -1,9 +1,9 @@
 /* This file is part of GNU cflow
-   Copyright (C) 1997, 2005, 2006 Sergey Poznyakoff
+   Copyright (C) 1997, 2005, 2006, 2007, 2009, 2010, 2011 Sergey Poznyakoff
  
    GNU cflow is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
  
    GNU cflow is distributed in the hope that it will be useful,
@@ -30,20 +30,19 @@ typedef struct {
 void parse_declaration(Ident*, int);
 void parse_variable_declaration(Ident*, int);
 void parse_function_declaration(Ident*, int);
-void parse_dcl(Ident*);
+void parse_dcl(Ident*, int maybe_knr);
 void parse_knr_dcl(Ident*);
 void parse_typedef();
 void expression();
 void initializer_list();
 void func_body();
-void declare(Ident*);
+void declare(Ident*, int maybe_knr);
 void declare_type(Ident*);
 int dcl(Ident*);
 int parmdcl(Ident*);
 int dirdcl(Ident*);
 void skip_struct();
 Symbol *get_symbol(char *name);
-Symbol *install_ident(char *name);
 void maybe_parm_list(int *parm_cnt_return);
     
 void call(char*, int);
@@ -85,6 +84,9 @@ print_token(TOKSTK *tokptr)
      case WORD:
      case MODIFIER:
      case STRUCT:
+     case PARM_WRAPPER:
+     case QUALIFIER:
+     case OP:
 	  fprintf(stderr, "`%s'", tokptr->token);
 	  break;
      case LBRACE0:
@@ -104,8 +106,8 @@ print_token(TOKSTK *tokptr)
      case TYPEDEF:
 	  fprintf(stderr, "`typedef'");
 	  break;
-     case OP:
-	  fprintf(stderr, "OP"); /* ouch!!! */
+     case STRING:
+	  fprintf(stderr, "\"%s\"", tokptr->token);
 	  break;
      default:
 	  fprintf(stderr, "`%c'", tokptr->type);
@@ -113,12 +115,12 @@ print_token(TOKSTK *tokptr)
 }
 
 static void
-file_error(char *msg, int near)
+file_error(char *msg, TOKSTK *tokptr)
 {
     fprintf(stderr, "%s:%d: %s", filename, tok.line, msg);
-    if (near) {
+    if (tokptr) {
 	fprintf(stderr, _(" near "));
-	print_token(&tok);
+	print_token(tokptr);
     }
     fprintf(stderr, "\n");
 }
@@ -215,6 +217,7 @@ save_token(TOKSTK *tokptr)
      case STRUCT:
      case PARM_WRAPPER:
      case WORD:
+     case QUALIFIER:
 	  if (need_space) 
 	       obstack_1grow(&text_stk, ' ');
 	  len = strlen(tokptr->token);
@@ -234,13 +237,33 @@ save_token(TOKSTK *tokptr)
      case EXTERN: /* storage class specifiers are already taken care of */
      case STATIC:
 	  break;
+     case ',':
+	  obstack_1grow(&text_stk, ',');
+	  need_space = 1;
+	  break;
      case '(':
 	  if (need_space) 
 	       obstack_1grow(&text_stk, ' ');
-	  /*FALLTHRU*/
-     default:
 	  obstack_1grow(&text_stk, tokptr->type);
 	  need_space = 0;
+	  break;
+     case ')':
+	  obstack_1grow(&text_stk, tokptr->type);
+	  need_space = 1;
+	  break;
+     case '[':
+     case ']':
+	  obstack_1grow(&text_stk, tokptr->type);
+	  need_space = 0;
+	  break;
+     case OP:
+	  obstack_1grow(&text_stk, ' ');
+	  obstack_grow(&text_stk, tokptr->token, strlen(tokptr->token));
+	  need_space = 1;
+	  break;
+     default:
+	  if (verbose)
+	       file_error(_("unrecognized definition"), tokptr);
      }
 }
 
@@ -305,6 +328,33 @@ skip_to(int c)
      }
 }
 
+int
+skip_balanced(int open_tok, int close_tok, int level)
+{
+     if (level == 0) {
+	  if (nexttoken() != open_tok) {
+	       return 1;
+	  }
+	  level++;
+     }
+     while (nexttoken()) {
+	  if (tok.type == LBRACE0 && open_tok == '{')
+	       tok.type = '{';
+	  else if (tok.type == RBRACE0 && close_tok == '}')
+	       tok.type = '}';
+	  
+	  if (tok.type == open_tok) 
+	       level++;
+	  else if (tok.type == close_tok) {
+	       if (--level == 0) {
+		    nexttoken();
+		    return 0;
+	       }
+	  } 
+     }
+     return -1;
+}
+
 int
 yyparse()
 {
@@ -318,6 +368,8 @@ yyparse()
 	  switch (tok.type) {
 	  case 0:
 	       return 0;
+	  case QUALIFIER:
+	       continue;
 	  case TYPEDEF:
 	       parse_typedef();
 	       break;
@@ -335,7 +387,7 @@ yyparse()
 	  }
 	  cleanup_stack();
      }
-     /*NOTREACHED*/
+     return 0;
 }
 
 static int
@@ -345,17 +397,29 @@ is_function()
      int res = 0;
 
      mark(sp);
-/*    if (tok.type == STRUCT)
-	nexttoken();*/
-     while (tok.type == TYPE ||
-	    tok.type == IDENTIFIER ||
-	    tok.type == MODIFIER ||
-	    tok.type == STATIC ||
-	    tok.type == EXTERN)
-	  nexttoken();
-     
-     if (tok.type == '(') 
-	  res = nexttoken() != MODIFIER;
+     while (1) {
+	  switch (tok.type) {
+	  case QUALIFIER:
+	  case TYPE:
+	  case IDENTIFIER:
+	  case MODIFIER:
+	  case STATIC:
+	  case EXTERN:
+	       nexttoken();
+	       continue;
+	  case PARM_WRAPPER:
+	       if (skip_balanced('(', ')', 0) == -1)
+		    file_error(_("unexpected end of file in declaration"),
+			       NULL);
+	       continue;
+	  case '(':
+	       res = nexttoken() != MODIFIER;
+	       break;
+	  default:
+	       break;
+	  }
+	  break;
+     }
      
      restore(sp);
      return res;
@@ -396,7 +460,8 @@ expression()
 	       break;
 	  case 0:
 	       if (verbose)
-		    file_error(_("unexpected end of file in expression"), 0);
+		    file_error(_("unexpected end of file in expression"),
+			       NULL);
 	       return;
 	    
 	  case IDENTIFIER:
@@ -451,7 +516,7 @@ parse_function_declaration(Ident *ident, int parm)
 	       nexttoken();
 	  else {
 	       if (verbose) 
-		    file_error(_("expected `;'"), 1);
+		    file_error(_("expected `;'"), &tok);
 	       error_recovery = 1;
 	  }
 	  goto restart;
@@ -468,7 +533,7 @@ parse_function_declaration(Ident *ident, int parm)
 	  break;
      case 0:
 	  if (verbose)
-	       file_error(_("unexpected end of file in declaration"), 0);
+	       file_error(_("unexpected end of file in declaration"), NULL);
      }
 }
 
@@ -498,10 +563,10 @@ fake_struct(Ident *ident)
 		    tos = curs + 1;
 	       }
 	       tokpush(hold.type, hold.line, hold.token);
-	  } else {
-	       if (tok.type != ';')
-		    file_error(_("missing `;' after struct declaration"), 0);
-	  }
+	  } else if (tok.type == '(')
+	       return 0;
+	  else if (tok.type != ';')
+	       file_error(_("missing `;' after struct declaration"), &tok);
 	  return 1;
      }
      return 0;
@@ -520,6 +585,8 @@ parse_variable_declaration(Ident *ident, int parm)
 	  }
 	  putback();
 	  skip_struct();
+	  while (tok.type == MODIFIER || tok.type == QUALIFIER)
+	       nexttoken();
 	  if (tok.type == IDENTIFIER) {
 	       TOKSTK hold = tok;
 	       restore(sp);
@@ -540,7 +607,7 @@ parse_variable_declaration(Ident *ident, int parm)
 	  }
      }
  again:
-     parse_dcl(ident);
+     parse_dcl(ident, 0);
      
  select:    
      switch (tok.type) {
@@ -550,7 +617,7 @@ parse_variable_declaration(Ident *ident, int parm)
 	  /*FALLTHROUGH*/
      default:
 	  if (verbose) 
-	       file_error(_("expected `;'"), 1);
+	       file_error(_("expected `;'"), &tok);
 	  /* FIXME: should putback() here */
 	  /* FALLTHRU */
      case ';':
@@ -575,7 +642,7 @@ parse_variable_declaration(Ident *ident, int parm)
 	  break;
      case 0:
 	  if (verbose)
-	       file_error(_("unexpected end of file in declaration"), 0);
+	       file_error(_("unexpected end of file in declaration"), NULL);
      }
 }
 
@@ -597,7 +664,8 @@ initializer_list()
 	       }
 	       break;
 	  case 0:
-	       file_error(_("unexpected end of file in initializer list"), 0);
+	       file_error(_("unexpected end of file in initializer list"),
+			  NULL);
 	       return;
 	  case ',':
 	       break;
@@ -613,86 +681,27 @@ void
 parse_knr_dcl(Ident *ident)
 {
      ident->type_end = -1;
-     parse_dcl(ident);
-     if (strict_ansi)
-	  return;
-     
-     switch (tok.type) {
-     case IDENTIFIER:
-     case TYPE:
-     case STRUCT:
-	  if (ident->parmcnt >= 0) {
-	       /* maybe K&R function definition */
-	       int parmcnt, stop;
-	       Stackpos sp, new_sp;
-	       Ident id;
-	       
-	       mark(sp);
-	       parmcnt = 0;
-	       
-	       for (stop = 0; !stop && parmcnt < ident->parmcnt;
-		    nexttoken()) {
-		    id.type_end = -1;
-		    switch (tok.type) {
-		    case LBRACE:
-		    case LBRACE0:
-			 putback();
-			 stop = 1;
-			 break;
-		    case TYPE:
-		    case IDENTIFIER:
-		    case STRUCT:
-			 putback();
-			 mark(new_sp);
-			 if (dcl(&id) == 0) {
-			      parmcnt++;
-			      if (tok.type == ',') {
-				   do {
-					tos = id.type_end; /* ouch! */
-					restore(new_sp);
-					dcl(&id);
-				   } while (tok.type == ',');
-			      } else if (tok.type != ';')
-				   putback();
-			      break;
-			 }
-			 /* else */
-			 /* FALLTHRU */
-		    default:
-			 restore(sp);
-			 return;
-		    }
-	       }
-	  }
-     }
+     parse_dcl(ident, !strict_ansi);     
 }
 
 void
 skip_struct()
 {
-     int lev = 0;
-     
      if (nexttoken() == IDENTIFIER) {
 	  nexttoken();
      } else if (tok.type == ';')
 	  return;
      
      if (tok.type == LBRACE || tok.type == LBRACE0) {
-	  do {
-	       switch (tok.type) {
-	       case 0:
-		    file_error(_("unexpected end of file in struct"), 0);
-		    return;
-	       case LBRACE:
-	       case LBRACE0:
-		    lev++;
-		    break;
-	       case RBRACE:
-	       case RBRACE0:
-		    lev--;
-	       }
-	       nexttoken();
-	  } while (lev);
+	  if (skip_balanced('{', '}', 1) == -1) {
+	       file_error(_("unexpected end of file in struct"), NULL);
+	       return;
+	  }
+     }
+
+     while (tok.type == PARM_WRAPPER) {
+	  if (skip_balanced('(', ')', 0) == -1)
+	      file_error(_("unexpected end of file in struct"), NULL);
      }
 }
 
@@ -717,7 +726,7 @@ parse_typedef()
 }
 
 void
-parse_dcl(Ident *ident)
+parse_dcl(Ident *ident, int maybe_knr)
 {
      ident->parmcnt = -1;
      ident->name = NULL;
@@ -725,7 +734,7 @@ parse_dcl(Ident *ident)
      dcl(ident);
      save_stack();
      if (ident->name)
-	  declare(ident);
+	  declare(ident, maybe_knr);
      else 
 	  undo_save_stack();
 }
@@ -733,22 +742,27 @@ parse_dcl(Ident *ident)
 int
 dcl(Ident *idptr)
 {
-     int type;
-
      while (nexttoken() != 0 && tok.type != '(') {
 	  if (tok.type == MODIFIER) {
 	       if (idptr && idptr->type_end == -1)
 		    idptr->type_end = curs-1;
+	  } else if (tok.type == PARM_WRAPPER) {
+	       if (skip_balanced('(', ')', 0) == -1) {
+		    file_error(_("unexpected end of file in function declaration"),
+			       NULL);
+		    return 1;
+	       }
 	  } else if (tok.type == IDENTIFIER) {
+	       int type;
+	       
 	       while (tok.type == IDENTIFIER)
 		    nexttoken();
 	       type = tok.type;
 	       putback();
-	       if (type == TYPE)
-		    continue;
-	       else if (type != MODIFIER) 
+	       if (!(type == TYPE ||
+		     type == MODIFIER || type == QUALIFIER))
 		    break;
-	  } else if (tok.type == ')') {
+	  } else if (tok.type == ')' || tok.type == ';') {
 	       return 1;
 	  }
      }
@@ -766,7 +780,7 @@ dirdcl(Ident *idptr)
      if (tok.type == '(') {
 	  dcl(idptr);
 	  if (tok.type != ')' && verbose) {
-	       file_error(_("expected `)'"), 1);
+	       file_error(_("expected `)'"), &tok);
 	       return 1;
 	  }
      } else if (tok.type == IDENTIFIER) {
@@ -789,7 +803,7 @@ dirdcl(Ident *idptr)
 	  else {
 	       maybe_parm_list(parm_ptr);
 	       if (tok.type != ')' && verbose) {
-		    file_error(_("expected `)'"), 1);
+		    file_error(_("expected `)'"), &tok);
 		    return 1;
 	       }
 	  }
@@ -797,26 +811,12 @@ dirdcl(Ident *idptr)
      if (wrapper)
 	  nexttoken(); /* read ')' */
 
-     if (tok.type == PARM_WRAPPER) {
-	  if (nexttoken() == '(') {
-	       int level = 0;
-	       while (nexttoken()) {
-		    if (tok.type == 0) {
-			 file_error(_("unexpected end of file in function declaration"),
-				    0);
-			 return 1;
-		    } else if (tok.type == '(') 
-			 level++;
-		    else if (tok.type == ')') {
-			 if (level-- == 0) {
-			      nexttoken();
-			      break;
-			 }
-		    } 
-	       }
-	  } else
-	       putback();
+     while (tok.type == PARM_WRAPPER) {
+	  if (skip_balanced('(', ')', 0) == -1)
+	       file_error(_("unexpected end of file in function declaration"),
+			  NULL);
      }
+     
      return 0;
 }
 
@@ -862,10 +862,11 @@ maybe_parm_list(int *parm_cnt_return)
 	       return;
 	  case ',':
 	       break;
+	  case QUALIFIER:
 	  case IDENTIFIER:
+	  case MODIFIER: /* unsigned * */
 	  case STRUCT:
 	  case UNION:
-	  case ENUM:
 	  case TYPE:
 	       parmcnt++;
 	       ident.storage = AutoStorage;
@@ -874,7 +875,8 @@ maybe_parm_list(int *parm_cnt_return)
 	       break;
 	  default:
 	       if (verbose)
-		    file_error(_("unexpected token in parameter list"), 1);
+		    file_error(_("unexpected token in parameter list"),
+			       &tok);
 	       level = 0;
 	       do {
 		    if (tok.type == '(') 
@@ -889,7 +891,7 @@ maybe_parm_list(int *parm_cnt_return)
 	  }
      }
      if (verbose)
-	  file_error(_("unexpected end of file in parameter list"), 0);
+	  file_error(_("unexpected end of file in parameter list"), NULL);
 }
 
 void
@@ -927,7 +929,7 @@ func_body()
 	  case RBRACE0:
 	       if (use_indentation) {
 		    if (verbose && level != 1)
-			 file_error(_("forced function body close"), 0);
+			 file_error(_("forced function body close"), NULL);
 		    for ( ; level; level--) {
 			 delete_autos(level);
 		    }
@@ -941,22 +943,76 @@ func_body()
 	       break;
 	  case 0:
 	       if (verbose)
-		    file_error(_("unexpected end of file in function body"), 0);
+		    file_error(_("unexpected end of file in function body"),
+			       NULL);
+	       caller = NULL;
 	       return;
 	  }
      }
+     caller = NULL;
+}
+
+int
+get_knr_args(Ident *ident)
+{
+     int parmcnt, stop;
+     Stackpos sp, new_sp;
+     Ident id;
+
+     switch (tok.type) {
+     case IDENTIFIER:
+     case TYPE:
+     case STRUCT:
+	  /* maybe K&R function definition */
+	  
+	  mark(sp);
+	  parmcnt = 0;
+	  
+	  for (stop = 0; !stop && parmcnt < ident->parmcnt;
+	       nexttoken()) {
+	       id.type_end = -1;
+	       switch (tok.type) {
+	       case LBRACE:
+	       case LBRACE0:
+		    putback();
+		    stop = 1;
+		    break;
+	       case TYPE:
+	       case IDENTIFIER:
+	       case STRUCT:
+		    putback();
+		    mark(new_sp);
+		    if (dcl(&id) == 0) {
+			 parmcnt++;
+			 if (tok.type == ',') {
+			      do {
+				   tos = id.type_end; /* ouch! */
+				   restore(new_sp);
+				   dcl(&id);
+			      } while (tok.type == ',');
+			 } else if (tok.type != ';')
+			      putback();
+			 break;
+		    }
+		    /* else */
+		    /* FALLTHRU */
+	       default:
+		    restore(sp);
+		    return 1;
+	       }
+	  }
+     }
+     return 0;
 }
 
 void
-declare(Ident *ident)
+declare(Ident *ident, int maybe_knr)
 {
      Symbol *sp;
      
      if (ident->storage == AutoStorage) {
 	  undo_save_stack();
-	  sp = install(ident->name);
-	  sp->type = SymIdentifier;
-	  sp->storage = ident->storage;
+	  sp = install_ident(ident->name, ident->storage);
 	  if (parm_level) {
 	       sp->level = parm_level;
 	       sp->flag = symbol_parm;
@@ -965,8 +1021,11 @@ declare(Ident *ident)
 	  sp->arity = -1;
 	  return;
      } 
-     
-     if ((ident->parmcnt >= 0 && !(tok.type == LBRACE || tok.type == LBRACE0))
+
+     if ((ident->parmcnt >= 0
+	  && (!maybe_knr || get_knr_args(ident) == 0)
+	  && !(tok.type == LBRACE || tok.type == LBRACE0 || tok.type == TYPE
+	       || tok.type == PARM_WRAPPER))
 	 || (ident->parmcnt < 0 && ident->storage == ExplicitExternStorage)) {
 	  undo_save_stack();
 	  /* add_external()?? */
@@ -977,11 +1036,16 @@ declare(Ident *ident)
      if (sp->source) {
 	  if (ident->storage == StaticStorage
 	      && (sp->storage != StaticStorage || level > 0)) {
-	       sp = install_ident(ident->name);
+	       sp = install_ident(ident->name, ident->storage);
 	  } else {
-	       error_at_line(0, 0, filename, ident->line, 
-			     _("%s/%d redefined"),
-			     ident->name, sp->arity);
+	       if (sp->arity >= 0)
+		    error_at_line(0, 0, filename, ident->line, 
+				  _("%s/%d redefined"),
+				  ident->name, sp->arity);
+	       else
+		    error_at_line(0, 0, filename, ident->line, 
+				  _("%s redefined"),
+				  ident->name);
 	       error_at_line(0, 0, sp->source, sp->def_line,
 			     _("this is the place of previous definition"));
 	  }
@@ -989,8 +1053,9 @@ declare(Ident *ident)
 
      sp->type = SymIdentifier;
      sp->arity = ident->parmcnt;
-     sp->storage = (ident->storage == ExplicitExternStorage) ?
-	  ExternStorage : ident->storage;
+     ident_change_storage(sp, 
+			  (ident->storage == ExplicitExternStorage) ?
+			  ExternStorage : ident->storage);
      sp->decl = finish_save_stack(ident->name);
      sp->source = filename;
      sp->def_line = ident->line;
@@ -1014,43 +1079,22 @@ declare_type(Ident *ident)
 	  if (sp->type == SymToken && sp->token_type == TYPE)
 	       break;
      if (!sp)
-	  sp = install(ident->name);
+	  sp = install(ident->name, INSTALL_UNIT_LOCAL);
      sp->type = SymToken;
      sp->token_type = TYPE;
      sp->source = filename;
      sp->def_line = ident->line;
      sp->ref_line = NULL;
      if (debug)
-	  printf(_("%s:%d: type %s\n"),
-		 filename,
-		 line_num,
-		 ident->name);
-}
-
-Symbol *
-install_ident(char *name)
-{
-     Symbol *sp;
-
-     sp = install(name);
-     sp->type = SymIdentifier;
-     sp->arity = -1;
-     sp->storage = ExternStorage;
-     sp->decl = NULL;
-     sp->source = NULL;
-     sp->def_line = -1;
-     sp->ref_line = NULL;
-     sp->caller = sp->callee = NULL;
-     sp->level = -1;
-     return sp;
+	  printf(_("%s:%d: type %s\n"), filename, line_num, ident->name);
 }
 
 Symbol *
 get_symbol(char *name)
 {
-     Symbol *sp;
+     Symbol *sp = lookup(name) ;
      
-     if (sp = lookup(name)) {
+     if (sp) {
 	  for (; sp; sp = sp->next) {
 	       if (sp->type == SymIdentifier && strcmp(sp->name, name) == 0)
 		    break;
@@ -1058,7 +1102,7 @@ get_symbol(char *name)
 	  if (sp)
 	       return sp;
      }
-     return install_ident(name);
+     return install_ident(name, ExternStorage);
 }
 
 Symbol *
@@ -1073,7 +1117,9 @@ add_reference(char *name, int line)
      refptr = xmalloc(sizeof(*refptr));
      refptr->source = filename;
      refptr->line = line;
-     append_to_list(&sp->ref_line, refptr);
+     if (!sp->ref_line)
+	  sp->ref_line = linked_list_create(free);
+     linked_list_append(&sp->ref_line, refptr);
      return sp;
 }
 
@@ -1089,10 +1135,10 @@ call(char *name, int line)
      if (sp->arity < 0)
 	  sp->arity = 0;
      if (caller) {
-	  if (!symbol_in_list(caller, sp->caller))
-	       append_to_list(&sp->caller, caller);
-	  if (!symbol_in_list(sp, caller->callee))
-	       append_to_list(&caller->callee, sp);
+	  if (!data_in_list(caller, sp->caller))
+	       linked_list_append(&sp->caller, caller);
+	  if (!data_in_list(sp, caller->callee))
+	       linked_list_append(&caller->callee, sp);
      }
 }
 
@@ -1103,10 +1149,10 @@ reference(char *name, int line)
      if (!sp)
 	  return;
      if (caller) {
-	  if (!symbol_in_list(caller, sp->caller))
-	       append_to_list(&sp->caller, caller);
-	  if (!symbol_in_list(sp, caller->callee))
-	       append_to_list(&caller->callee, sp);
+	  if (!data_in_list(caller, sp->caller))
+	       linked_list_append(&sp->caller, caller);
+	  if (!data_in_list(sp, caller->callee))
+	       linked_list_append(&caller->callee, sp);
      }
 }
 
