@@ -1,20 +1,20 @@
-/*  $Id: parser.h,v 1.5 2005/05/18 11:36:44 gray Exp $
- *  cflow
- *  Copyright (C) 1997 Gray
+/*  This file is part of GNU cflow
+ *  Copyright (C) 1997, 2007, 2010 Sergey Poznyakoff
  *  
- *  This program is free software; you can redistribute it and/or modify
+ *  GNU cflow is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
+ *  the Free Software Foundation; either version 3 of the License, or
  *  (at your option) any later version.
  *  
- *  This program is distributed in the hope that it will be useful,
+ *  GNU cflow is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  *  GNU General Public License for more details.
  *  
  *  You should have received a copy of the GNU General Public License
- *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *  along with GNU cflow; if not, write to the Free Software
+ *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
+ *  MA 02110-1301 USA
  */
 /* tokens */
 #define WORD 257
@@ -35,6 +35,7 @@
 #define TYPE 270
 #define STRING 271
 #define PARM_WRAPPER 272
+#define QUALIFIER 273
 
 typedef struct {
     char *str;
